import { cv } from "../../cv";
import { convertImage } from "../common/processors";
import { FaceBox, FaceDetectOption, FaceDetector, nms } from "./common";

const MODEL_URL_CONFIG = {
	YOLOV5S_ONNX: `https://www.modelscope.cn/models/luyizhi/basic_cv/resolve/master/facedet/yolov5s.onnx`,
	YOLOV5S_MNN: `https://www.modelscope.cn/models/luyizhi/basic_cv/resolve/master/facedet/yolov5s.mnn`,
};

export class Yolov5Face extends FaceDetector {

	public static async load(type?: keyof typeof MODEL_URL_CONFIG) {
		return this.cacheModel(MODEL_URL_CONFIG[type ?? "YOLOV5S_ONNX"], { createModel: true }).then(r => r.model);
	}

	public async doPredict(image: cv.Mat, option?: FaceDetectOption): Promise<FaceBox[]> {
		const input = this.input;
		const resizedImage = image.resize(input.shape[2], input.shape[3]);
		const ratioWidth = image.width / resizedImage.width;
		const ratioHeight = image.height / resizedImage.height;
		const nchwImageData = convertImage(resizedImage.data, { sourceImageFormat: "bgr", targetColorFormat: "bgr", targetShapeFormat: "nchw", targetNormalize: { mean: [127.5], std: [127.5] } });

		const outputData = await this.session.run({ input: nchwImageData }).then(r => r.output);
		const outShape = this.outputs["output"].shape;

		const faces: FaceBox[] = [];
		const threshold = option?.threshold ?? 0.5;
		for (let i = 0; i < outShape[1]; i++) {
			const beg = i * outShape[2];
			const rectData = outputData.slice(beg, beg + outShape[2]);
			const x = parseInt(rectData[0] * ratioWidth as any);
			const y = parseInt(rectData[1] * ratioHeight as any);
			const w = parseInt(rectData[2] * ratioWidth as any);
			const h = parseInt(rectData[3] * ratioHeight as any);
			const score = rectData[4] * rectData[15];
			if (score < threshold) continue;
			faces.push(new FaceBox({
				x1: x - w / 2, y1: y - h / 2,
				x2: x + w / 2, y2: y + h / 2,
				score, imw: image.width, imh: image.height,
			}))
		}
		return nms(faces, option?.mnsThreshold ?? 0.3).map(box => box.toInt());
	}

}
